/*

Copyright 2018 Surface Concept GmbH, Germany

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.

*/

#include "stdafx.h"
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <thread>
#include <chrono>
#include <time.h>

#include <scTDC.h>
#include <scTDC_error_codes.h>


typedef uint64_t EVENT_BUF_UNITTYPE;

struct myglobals {
	int device_descriptor = 0;
	int nr_measurements = 1000;
	int measurement_counter = 1;
	int acqtime_ms = 100;
	bool measurements_finished = false;
	bool abort_program = false;
	double binsize_time = 0.5;
	std::ofstream* outf_ptr = nullptr; // pointer to the text file stream object
} G;



void handle_end_of_measure(void *priv) {
	if (G.measurement_counter < G.nr_measurements) {
		G.measurement_counter++;
		std::cout << "Starting measurement #" << std::setw(4) << G.measurement_counter << std::endl;
		// -----------------------------------
		// ---    start next measurement   ---
		// -----------------------------------
		int retcode = ::sc_tdc_start_measure2(G.device_descriptor, G.acqtime_ms);
		if (retcode < 0) {
			std::cerr << "sc_tdc_start_measure2 returned " << retcode << std::endl;
			char errmsg[ERRSTRLEN];
			::sc_get_err_msg(retcode, errmsg);
			std::cerr << errmsg << std::endl;
			G.abort_program = true;
		}
		return;
	}
	else if (G.measurement_counter >= G.nr_measurements) {
		G.measurements_finished = true;
	}
}

void handle_dld_event(void *priv, const struct sc_DldEvent *const event_array, size_t event_array_len) {
	myglobals& g = *reinterpret_cast<myglobals*>(priv); // for demo purposes, get our private data from priv
	// in this program, we could also directly access the global variable G
	if (g.outf_ptr == nullptr) // safety check
		return;
	std::ofstream& outf = *(g.outf_ptr);

	// output values of the event data fields into file (one row per event)
	std::ostringstream oss;
	for (size_t i = 0; i < event_array_len; i++) {
		oss << std::setw(8) << event_array[i].dif1 // x coordinate
			<< std::setw(8) << event_array[i].dif2 // y coordinate
			<< std::setw(30) << std::setprecision(std::numeric_limits<double>::digits10 + 1)
			<< (static_cast<double>(event_array[i].sum) * G.binsize_time * 1e-6)   // time (1e-6 converts from ns to ms)
			<< std::setw(8) << event_array[i].adc  // analog-digital converter value
			<< std::setw(8) << event_array[i].signal1bit // "state" input
			<< std::setw(8) << event_array[i].master_rst_counter // master reset counter
			<< std::endl;
	}
	outf << oss.str();
	//outf.flush();
}

void handle_start_of_measurement(void *priv) {
}

void handle_millisecond_countup(void *priv) {
}

void handle_statistics(void *priv, const struct statistics_t *statdata) {
}

void handle_tdc_event(void *priv, const struct sc_TdcEvent *const event_array, size_t event_array_len) {
}

std::string end_of_measurement_message() {
	std::ostringstream oss;
	std::chrono::system_clock::time_point tnow_tmp = std::chrono::system_clock::now();
	auto duration = tnow_tmp.time_since_epoch();
	auto millis = std::chrono::duration_cast<std::chrono::milliseconds>(duration).count();
	millis = millis % 1000;
	time_t tnow_tmp2 = std::chrono::system_clock::to_time_t(tnow_tmp);
	struct tm tnow;
	localtime_s(&tnow, &tnow_tmp2);
	oss << "# END OF MEASUREMENT   DATE & TIME : " <<
		std::setw(4) << std::setfill('0') << tnow.tm_year + 1900 << "-" <<
		std::setw(2) << std::setfill('0') << tnow.tm_mon + 1 << "-" <<
		std::setw(2) << std::setfill('0') << tnow.tm_mday << " " <<
		std::setw(2) << std::setfill('0') << tnow.tm_hour << ":" <<
		std::setw(2) << std::setfill('0') << tnow.tm_min << ":" <<
		std::setw(2) << std::setfill('0') << tnow.tm_sec << " (+" <<
		std::setw(3) << std::setfill('0') << millis << "ms)" << std::endl << std::endl;
	return oss.str();
}



/************************************************************************
 * This function is the main routine for the TDC-related action         *
 ************************************************************************/
int write_tdc_events_to_textfile(std::ofstream& outf, long acqtime_ms_arg, long nrmeas_arg) {
	// ---------------------------------------------------------------
	// ---               initialize device                         ---
	// ---------------------------------------------------------------
	int retcode = ::sc_tdc_init_inifile(R"(tdc_gpx3.ini)"); // R"()" is a raw string literal: backslashes are not interpreted as escape characters
	if (retcode < 0) {
		std::cout << "Init Inifile returned " << retcode << std::endl;
		char errmsg[ERRSTRLEN];
		::sc_get_err_msg(retcode, errmsg);
		std::cout << errmsg << std::endl;
		return retcode;
	}
	G.device_descriptor = retcode;
	G.acqtime_ms = acqtime_ms_arg;
	G.nr_measurements = nrmeas_arg;
	G.outf_ptr = &outf;

	// ---------------------------------------------------------------
	// ---        get time bin size in nanoseconds                 ---
	// ---------------------------------------------------------------
	::sc_tdc_get_binsize2(G.device_descriptor, &G.binsize_time); // base TDC time bin size
	// for 2D DLD, this should show 0.00685871 ns and then, time (ms) is retrieved correctly
	std::cout << "# Time pixel corresponds to " << G.binsize_time << " ns" << std::endl;
	// ---------------------------------------------------------------
	// ---     open a pipe for receiving events via callbacks      ---
	// ---  1. fill a structure with function pointers to our      ---
	// ---     callback handlers                                   ---
	// ---  2. open the pipe                                       ---
	// ---------------------------------------------------------------
	::sc_pipe_callback_params_t cb_params;
	cb_params.callbacks = new sc_pipe_callbacks;
	cb_params.callbacks->priv = reinterpret_cast<void*>(&G); // use this field to transport private data into callbacks
	cb_params.callbacks->end_of_measure = handle_end_of_measure;
	cb_params.callbacks->dld_event = handle_dld_event;
	cb_params.callbacks->millisecond_countup = handle_millisecond_countup; // unused callbacks:
	cb_params.callbacks->start_of_measure = handle_start_of_measurement;    // do not leave 
	cb_params.callbacks->statistics = handle_statistics;          // function pointers 
	cb_params.callbacks->tdc_event = handle_tdc_event;           // uninitialized
	::sc_pipe_open2(G.device_descriptor, USER_CALLBACKS, reinterpret_cast<void*>(&cb_params));
	int retcode3 = ::sc_tdc_pipe_open2(G.device_descriptor, 1 << 24, nullptr, nullptr);
	if (retcode3 < 0)  {
		std::cout << "sc_tdc_pipe_open2 returned " << retcode3 << std::endl;
		char errmsg[ERRSTRLEN];
		::sc_get_err_msg(retcode3, errmsg);
		std::cout << errmsg << std::endl;
		return retcode3;
	}
	// ---------------------------------------------------------------
	// ---       write header into the output file                 ---
	// ---------------------------------------------------------------
	outf << "#" << std::setw(7) << "x" // x coordinate
		<< std::setw(8) << "y" // y coordinate
		<< std::setw(30) << "time (ms)"
		<< std::setw(8) << "ADC"
		<< std::setw(8) << "state"
		<< std::setw(8) << "MRCtr"
		<< std::endl;
	// ---------------------------------------------------------------
	// ---      start first exposure ("measurement")               ---
	// ---------------------------------------------------------------
	std::cout << "Starting measurement #" << std::setw(4) << G.measurement_counter << std::endl;
	int retcode4 = ::sc_tdc_start_measure2(G.device_descriptor, G.acqtime_ms);
	if (retcode4 < 0) {
		std::cout << "sc_tdc_start_measure2 returned " << retcode4 << std::endl;
		char errmsg[ERRSTRLEN];
		::sc_get_err_msg(retcode4, errmsg);
		std::cout << errmsg << std::endl;
		G.abort_program = true;
	}
	// ---------------------------------------------------------------
	// ---  iterate reading events until all exposures have        ---
	// ---  finished (start of subsequent exposures happens in     ---
	// ---  the "measurement complete" callback)                   ---
	// ---------------------------------------------------------------
	while (!G.measurements_finished && !G.abort_program) {
		std::this_thread::sleep_for(std::chrono::milliseconds(200)); // wait
	}
	std::this_thread::sleep_for(std::chrono::milliseconds(100)); // wait once more
	// ---------------------------------------------------------------
	// ---              release TDC                                ---
	// ---------------------------------------------------------------
	std::cout << end_of_measurement_message();
	::sc_tdc_deinit2(G.device_descriptor);
	return 0;
}

int _tmain(int argc, _TCHAR* argv[])
{
	// ----------------------------------------------------------------
	// ---     parse command line arguments here, call              ---
	// ---     write_tdc_events_to_textfile(...) to perform what    ---
	// ---     the program is aimed at                              ---
	// ----------------------------------------------------------------
	if (argc != 4 && argc != 5) { // program name itself is counted in argc
		std::cout << std::endl << "scTDC_events_to_txt_usercallbacks  -  Surface Concept GmbH" << std::endl;
		std::cout << "Version 02.10.2018" << std::endl << std::endl;
		std::cout << "Usage: " << std::endl
			<< "  scTDC_events_to_txt_usercallbacks.exe  [-f] OUTPUTFILE  ACQUISITION_TIME(ms)  NUMBER_OF_MEASUREMENTS" << std::endl
			<< "                            -f ... overwrite OUTPUTFILE if exists" << std::endl;
		exit(1);
	}
	// at this point, argc == 4 || argc == 5
	std::size_t argshift = 0; // if optional argument is used, shift the rest
	bool overwrite = false;
	std::wstring arg1{ argv[1] };
	if (arg1.compare(L"-f") == 0) {
		overwrite = true;
		argshift++;
	}

	if (overwrite && argc != 5) {
		std::cout << "arguments missing" << std::endl;
		exit(1);
	}

	std::wstring outfile_path{ argv[1 + argshift] };

	//  -----   test if file exists   -------------------------------------
	if (!overwrite) {
		bool outfile_exists = false;
		{	// scope for limited lifetime of ifstream object
			std::ifstream outf_read{ outfile_path };
			if (outf_read.is_open())
				outfile_exists = true;
			outf_read.close();
			if (outfile_exists) {
				std::cout << "Specified output file exists. Aborting." << std::endl;
				exit(2);
			}
		}
	}
	// ----------------------------------------------------------------------
	std::ofstream outf_write{ outfile_path };
	if (!outf_write.is_open()) {
		std::cout << "Specified output file cannot be opened. Aborting." << std::endl;
		exit(3);
	}
	else
		std::wcout << "# Event data will be written to : " << outfile_path << std::endl;

	// ----------------------------------------------------------------------
	long acqtime_ms_local = 0;
	bool acqtime_ms_ok = false;
	{
		std::basic_istringstream<_TCHAR> iss{ argv[2 + argshift] };
		iss >> acqtime_ms_local;
		if (!iss.fail()) {
			acqtime_ms_ok = true;
			std::cout << "# Acquisition time is " << acqtime_ms_local << " ms" << std::endl;
		}
		else {
			std::wcout << "Specified acquisition time " << argv[2 + argshift] << " cannot be parsed to an integer number" << std::endl;
			outf_write.close();
			exit(4);
		}
	}
	// --------------------------------------------------------------------------
	long number_of_measurements = 0;
	bool number_of_measurements_ok = false;
	{
		std::basic_istringstream<_TCHAR> iss(argv[3 + argshift]);
		iss >> number_of_measurements;
		if (!iss.fail()) {
			number_of_measurements_ok = true;
			std::cout << "# Number of measurements is " << number_of_measurements << std::endl;
		}
		else {
			std::wcout << "Specified number of measurements " << argv[3 + argshift] << " cannot be parsed to an integer number" << std::endl;
			outf_write.close();
			exit(5);
		}
	}
	// ---------------------------------------------------------------------------
	write_tdc_events_to_textfile(outf_write, acqtime_ms_local, number_of_measurements);

}

